#!/usr/bin/perl -T

#---------------------------------------------------------------------
# Copyright (C) 2003 Chad Redman <chad_eu2@katica.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#---------------------------------------------------------------------

# $Id: eventcheck.cgi,v 1.6 2005/08/27 03:26:57 chad Exp $

# File:     eventcheck.cgi
# Version:  0.4.1
# Author:   Chad Redman <chad_eu2@katica.org>

# This as a CGI script for doing event validations over the web. It can
# be run on a web server from any CGI-capable directory, as long as the EU2
# subdirectory (the one containing the Parse/*.pm files) resides in the same
# directory. The script requires the additional modules CGI and HTML::Parser
# (which seems to require HTML::Tagset) to be installed in the Perl site
# library directory.

# If you want to have this CGI script separate from the rest
# of the EU2Lint files, change this to the full path to the EU2Lint directory
use lib '.';

########### End of Configuration ###########

use EU2::Parse::EventParser;
use HTML::Entities();

$VERSION = '0.4.1';
$PARSERVERSION = EU2::Parse::EventParser->Version;

EU2::Parse::EventParser->ErrorFunc (sub {
    print STDOUT '<span class="error">* ', HTML::Entities::encode( join('', @_) ), "</span><br>\n";
});

#-------------------------------------

use CGI;
use CGI::Carp 'fatalsToBrowser';

$CGI::POST_MAX=1024 * 512;  # max 512K posts

$q = new CGI;
print $q->header;

$mode = $q->param('mode');

my $input;
if (!$mode) {
    print <<"EOF";
    <html>
    <head><title>EU2 Online Event Syntax Validator</title></head>
    <style>
      TEXTAREA { font-size: 8pt; font-family: monospace; }
      .code    {font-family: monospace; color: brown; font-size: 120%}
      ul.changelog {font-size: 80%; }
    </style>
    <body>
    <h2 align="center">EU2 Event Syntax Validator (version $VERSION, Event Parser
    version $PARSERVERSION)</h2>
    Part of the <a href="http://www.katica.org/cer28/eu2/EU2EventParser">The EU2 Event Validator Project</a><br><br>

    <strong>Change Log</strong>
    <ul class="changelog">
      <li>Parser Version 0.9.4: Country tag -6 (current emperor).
          Triggers: <span class="code">elector</span>, <span class="code">emperor</span>.
          Command actions: <span class="code">flagname</span>.
          Give warning for use of provinceculture trigger.
          Updated for new countries and cultures in EU2 v1.0.8.<br>
      </li>
  
      <li>
      Parser Version 0.9.3: Handle <span class="code">removecot</span> command action<br>
      </li>
  
      <li>
      Version 0.4: Use new parser engine. Added <span class="code">badboy</span> command action added to EU2 v1.07.
          Handle multi-line quoted strings. Detect accented characters as valid letters in a word. Maximum
          leader and monarch id increased from 20,000 to 200,000. Escape special HTML
          characters more rigorously. Report the number of events successfully processed.<br>
      </li>
  
      <li>
      Version 0.3.1: Removed warning when encountering the <span class="code">core</span>
      trigger, which wasn't a working trigger before 1.06. Added a check for the same event ID used more than once.<br>
      </li>
  
      <li>
      Version 0.3: Added the <span class="code">ai</span> and <span class="code">flag</span> triggers, and
      <span class="code">setflag</span>, <span class="code">clrflag</span>, and <span class="code">ai</span>
      commands available in 1.06.<br><br>
      </li>
    </ul>
    <form enctype="multipart/form-data" method="POST">
      <b>From a file</b>
      <input class="small" type="file" name="filename"><br>
      <b>Range warnings: </b>
      <input type="radio" name="range_check" value="strict" checked>strict
      <input type="radio" name="range_check" value="loose">loose
      <input type="radio" name="range_check" value="none">none
      <br>
      <input type="hidden" name="dup_id_check" value="1">
      <input type="hidden" name="mode" value="process_file">
      <input type="submit">
    </form>
    <hr>
    <form method="POST">
      <b>Or paste the code for one or more events below.</b><br>
      <b>Range warnings: </b>
      <input type="radio" name="range_check" value="strict" checked>strict
      <input type="radio" name="range_check" value="loose">loose
      <input type="radio" name="range_check" value="none">none
      <br>
      <input type="hidden" name="dup_id_check" value="1">
      <input type="submit"><br>
      <textarea name="txtEvent" rows=30 cols=80></textarea><br>
      <input type="hidden" name="mode" value="process">
    </form>
EOF

    exit 0;
} elsif ($mode eq 'process' or $mode eq 'process_file') {
    print <<EOF;
    <html><head>
    <style>
        .error {font-family: monospace; color: brown; margin-left: 2em;}
    </style>
    </head>
    <body>
    <i>retrieving data--Any errors found will be reported below... </i><br>
EOF

    $EU2::Parse::EventParser::opt_check_range = $q->param('range_check') || 'strict';
    $opt_check_dup_ids = $q->param('dup_id_check') || '';

    my $parser = new EU2::Parse::EventParser;

    if ($q->param('filename')) {   # tried to upload a file
        my $fh = $q->upload('filename');
        if ($fh) {
            $parser->fileHandle($fh);
        } else {
            print "<b>Failed to upload file</b>\n";
            exit 0;
        }
    } elsif ($q->param('txtEvent')) {
        $parser->source( $q->param('txtEvent') );
    } else {
        print "<b>Failed to get any input</b>\n";
        exit 0;
    }

    # Event properly initialized. Now start analyzing
    print "Started parse of ", $q->param('filename')||'pasted event', " at ", scalar(localtime),"<br>\n";
    print "<p>\n";
    $parser->parse;
    my %ids;
    foreach my $id (@{$parser->{ids}}) {
        if (defined $ids{$id}) {
            $opt_check_dup_ids and EU2::Parse::EventParser->Log("Duplicate Event ID $id");
        } else {
            $ids{$id} = 1;
        }
    }

    print "</p>\n";
    print scalar(@{$parser->events}), " event(s) successfully parsed from ", $q->param('filename')||'pasted text', " at ", scalar(localtime),"\n";

} else {
    print "Unknown mode: '$mode'";
    exit 0;
}
